import socket
import time
from netkit.contrib.tcp_client import TcpClient

from ..share import constants
from ..share.log import logger
from ..share.task import Task


class Connection(object):

    def __init__(self, worker, address, conn_timeout):
        self.worker = worker
        # 直接创建即可
        self.client = TcpClient(Task, address=address, timeout=conn_timeout)

    def run(self):
        while self.worker.enabled:
            try:
                self._handle()
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur. worker: %s',
                             self.worker, exc_info=True)

    def _handle(self):
        while self.worker.enabled and self.closed():
            if not self._connect():
                logger.error('connect fail. worker: %s, address: %s, sleep %ss',
                             self.worker, self.client.address, self.worker.app.reconnect_interval)
                time.sleep(self.worker.app.reconnect_interval)

        if not self.worker.enabled:
            # 安全退出
            return

        # 跟gateway要task
        self._ask_for_task()
        self._read_message()

    def _ask_for_task(self):
        task = Task()
        task.cmd = constants.CMD_WORKER_ASK_FOR_TASK
        task.room_id = self.worker.room_id

        return self.write(task.pack())

    def _connect(self):
        try:
            self.client.connect()
        except KeyboardInterrupt:
            raise
        except:
            return False
        else:
            self.worker.app.events.open_conn(self)
            for bp in self.worker.app.blueprints:
                bp.events.open_app_conn(self)

            return True

    def write(self, data):
        """
        发送数据    True: 成功   else: 失败
        """
        if self.client.closed():
            logger.error('connection closed. worker: %s, data: %r', self.worker, data)
            return False

        # 只支持字符串
        self.worker.app.events.before_response(self, data)
        for bp in self.worker.app.blueprints:
            bp.events.before_app_response(self, data)

        ret = self.client.write(data)
        if not ret:
            logger.error('connection write fail. worker: %s, data: %r', self.worker, data)

        for bp in self.worker.app.blueprints:
            bp.events.after_app_response(self, data, ret)
        self.worker.app.events.after_response(self, data, ret)

        return ret

    def _read_message(self):
        task = None

        while 1:
            try:
                # 读取数据 gw_box
                task = self.client.read()
            except socket.timeout:
                # 超时了
                if not self.worker.enabled:
                    return
                else:
                    # 继续读
                    continue
            else:
                # 正常收到数据了
                break

        if task:
            self._on_read_complete(task)

        if self.closed():
            self._on_connection_close()

    def _on_connection_close(self):
        # 链接被关闭的回调

        logger.error('connection closed. worker: %s, address: %s', self.worker, self.client.address)

        for bp in self.worker.app.blueprints:
            bp.events.close_app_conn(self)
        self.worker.app.events.close_conn(self)

    def _on_read_complete(self, task):
        """
        数据获取结束
        """
        self.worker.push_task(task)

    def close(self):
        """
        直接关闭连接
        """
        self.client.close()

    def closed(self):
        """
        连接是否已经关闭
        :return:
        """
        return self.client.closed()





NAME = 'pansy'

CMD_CLIENT_REQ              = 10  # 透传client请求
CMD_CLIENT_CLOSED           = 20  # 客户端连接被关闭

CMD_WRITE_TO_WORKER         = 100 # 给worker发送消息

CMD_WORKER_ASK_FOR_TASK     = 210 # 请求任务

CMD_WRITE_TO_USERS          = 230 # 主动下发
CMD_CLOSE_USERS             = 250 # 关闭多个客户端


DEFAULT_CONFIG = dict(
    HOST='127.0.0.1',
    PORT=9700,

    # 启动的room id 列表，闭区间
    ROOM_ID_BEGIN=None,
    ROOM_ID_END=None,
    # 同时并发房间数
    ROOM_CONCURRENT=1,

    DEBUG=False,

    BOX_CLASS='netkit.box.Box',

    # master class
    MASTER_CLASS='pansy.master.Master',

    # station class
    STATION_CLASS='pansy.station.Station',

    # worker class
    WORKER_CLASS='pansy.worker.Worker',

    NAME=NAME,

    CONN_TIMEOUT=None,

    # 重连等待时间。worker和client都在用
    RECONNECT_INTERVAL=1,

    STOP_TIMEOUT=None,

    # 处理task超时(秒). 超过后会打印fatal日志. None 代表永不超时
    WORK_TIMEOUT=None,

    # 等待处理的tasks的最大值。<=0代表无限
    TASKS_MAX_SIZE=-1,
)




import multiprocessing
import signal
import time
import _thread
import setproctitle


class Master:

    app = None

    processes = None

    enabled = True

    def __init__(self, app):
        self.app = app
        self.processes = list()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            'master:%s-%s' % (self.app.room_id_begin, self.app.room_id_end)
        ))
        self._handle_signals()

        self._spawn_stations()

    def _create_station(self, room_id_begin, room_id_end):
        process = multiprocessing.Process(
            target=self.app.station_class(self.app, room_id_begin, room_id_end).run,
        )
        # master不可以独自退出
        process.daemon = False
        # 标记room_id
        process.init_params = dict(
            room_id_begin=room_id_begin,
            room_id_end=room_id_end,
        )

        process.start()

        return process

    def _spawn_stations(self):
        """
        监控进程
        :return:
        """
        for room_id_begin in range(self.app.room_id_begin, self.app.room_id_end+1, self.app.room_concurrent):
            # 注意要-1，因为是闭区间
            room_id_end = min(room_id_begin + self.app.room_concurrent-1, self.app.room_id_end)
            self.processes.append(self._create_station(
                room_id_begin, room_id_end
            ))

        while True:
            for idx, process in enumerate(self.processes):
                if process and not process.is_alive():
                    self.processes[idx] = None

                    if self.enabled:
                        # 需要重启启动process
                        self.processes[idx] = self._create_station(**process.init_params)

            if not self.enabled and not any(self.processes):
                # 可以彻底停掉了
                break

            time.sleep(0.1)

    def _handle_signals(self):
        def kill_processes_later(processes, timeout):
            """
            等待一段时间后杀死所有进程
            :param processes:
            :param timeout:
            :return:
            """
            def _kill_processes():
                # 等待一段时间
                time.sleep(timeout)

                for p in processes:
                    if p and p.is_alive():
                        # 说明进程还活着
                        p.kill()

            _thread.start_new_thread(_kill_processes, ())

        def stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self.enabled = False

            # 一定要这样，否则后面kill的时候可能会kill错
            processes = self.processes[:]

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            for p in processes:
                if p:
                    p.terminate()

            if self.app.stop_timeout is not None:
                kill_processes_later(processes, self.app.stop_timeout)

        # INT为强制结束
        signal.signal(signal.SIGINT, stop_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, signal.SIG_IGN)





from netkit.box import Box
from collections import OrderedDict

# 如果header字段变化，那么格式也会变化
HEADER_ATTRS = OrderedDict([
    ('magic', ('i', 2037952207)),
    ('version', ('h', 0)),
    ('flags', ('h', 0)),
    ('packet_len', ('i', 0)),
    ('cmd', ('i', 0)),
    ('ret', ('i', 0)),
    ('client_ip_num', ('16s', b'')),
    ('uid', ('q', 0)),
    ('room_id', ('q', 0)),
    ('custom', ('q', 0)),
    ])


# 是否来自inner
FLAG_INNER = 0x01
# 是否IPV6
FLAG_IPV6 = 0x01 << 1


class Task(Box):
    header_attrs = HEADER_ATTRS

    @property
    def client_ip(self):
        """
        获取字符串格式的IP地址
        对端传过来的本身就是原始的字节序的二进制流，所以不需要做任何特殊转化，直接使用即可。
        :return:
        """
        import socket
        if self.ipv6:
            return socket.inet_ntop(socket.AF_INET6, self.client_ip_num)
        else:
            # ipv4只有4个字节
            return socket.inet_ntop(socket.AF_INET, self.client_ip_num[:4])

    @property
    def inner(self):
        """
        是否来自inner
        :return: True/False
        """
        return bool(self.flags & FLAG_INNER)

    @property
    def ipv6(self):
        """
        是否是ipv6
        :return: True/False
        """
        return bool(self.flags & FLAG_IPV6)



from setuptools import setup, find_packages

setup(
    name='pansy',
    version='2.0.16',
    zip_safe=False,
    platforms='any',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    python_requires='>=3',
    install_requires=['pycrypto', 'protobuf>=3.9.1', 'setproctitle', 'netkit', 'pyglet', 'events'],
    url='https://github.com/dantezhu/pansy',
    license='MIT',
    author='dantezhu',
    author_email='zny2008@gmail.com',
    description='pansy',
)
